// main.cpp

// Include necessary headers
#include <glad/glad.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>              // For 3D math (like matrices)
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

#include "imgui.h"
#include "imgui_impl_glfw.h"
#include "imgui_impl_opengl3.h"

#include <iostream>
#include <vector>
#include <map>
#include <sstream>
#include <tuple>
#include <string>

// Screen dimensions
unsigned int SCR_WIDTH = 1000;
unsigned int SCR_HEIGHT = 800;

// Sidebar dimensions
const unsigned int SIDEBAR_WIDTH = 250;

// Camera settings
glm::vec3 cameraPos   = glm::vec3(0.0f, 0.0f,  20.0f);
glm::vec3 cameraFront = glm::vec3(0.0f, 0.0f, -1.0f);
glm::vec3 cameraUp    = glm::vec3(0.0f, 1.0f,  0.0f);

float yaw   = -90.0f; // Yaw is initialized to -90.0 degrees to point towards negative Z
float pitch = 0.0f;

float lastX = SCR_WIDTH / 2.0f;
float lastY = SCR_HEIGHT / 2.0f;

float fov = 45.0f;

bool firstMouse = true;

// Timing
float deltaTime = 0.0f; // Time between current frame and last frame
float lastFrame = 0.0f; // Time of last frame

// Control Panel visibility
bool showUI = true;

// New global variables for your requests
bool showGridPoints = false; // Initially set to false

int sphereSizeInt = 1;      // Default sphere size
int lineThicknessInt = 3;   // Default line thickness
float sphereScale = 1.0f;   // Scaling factor for sphere size
float lineThickness = 3.0f; // Line thickness

// New global variables for JSON input
char jsonInput[1024 * 16] = "";
std::string jsonErrorMessage = "";

// Function prototypes
void framebuffer_size_callback(GLFWwindow* window, int width, int height);
void processInput(GLFWwindow *window);

void mouse_callback(GLFWwindow* window, double xpos, double ypos);
void scroll_callback(GLFWwindow* window, double xoffset, double yoffset);

unsigned int createShaderProgram(const char* vertexShaderSource, const char* fragmentShaderSource);

// Sphere generation
struct Sphere {
    glm::vec3 position;
    glm::vec3 color; // Gray or Red
};

void generateSphere(std::vector<float>& vertices, std::vector<unsigned int>& indices, float radius, unsigned int sectorCount, unsigned int stackCount);

void updateLines(const std::map<std::tuple<int, int, int>, Sphere>& spheres,
                std::vector<float>& lineVertices, unsigned int lineVBO);

// Key callback for toggling UI
void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods);

// Function to update window title based on mode
void updateWindowTitle(GLFWwindow* window, bool showUI);

// Helper function to convert world coordinates to screen coordinates
bool WorldToScreen(const glm::vec3& worldPos, const glm::mat4& view, const glm::mat4& projection, ImVec2& screenPos);

// Function to parse JSON input
bool parseHPModelJson(const std::string& jsonInput, std::map<std::tuple<int, int, int>, Sphere>& spheres, std::string& errorMessage);

// Entry point
int main()
{
    // Initialize GLFW
    if (!glfwInit())
    {
        std::cerr << "Failed to initialize GLFW" << std::endl;
        return -1;
    }

    // Set GLFW context version (OpenGL 3.3 Core)
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    // Create window
    GLFWwindow* window = glfwCreateWindow(SCR_WIDTH, SCR_HEIGHT, "3D Spheres with Connections - UI Interaction Mode", NULL, NULL);
    if (window == NULL)
    {
        std::cout << "Failed to create GLFW window" << std::endl;
        glfwTerminate();
        return -1;
    }
    glfwMakeContextCurrent(window);

    // Set callbacks
    glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
    // Mouse and scroll callbacks
    glfwSetCursorPosCallback(window, mouse_callback);
    glfwSetScrollCallback(window, scroll_callback);
    // Key callback
    glfwSetKeyCallback(window, key_callback);

    // Initialize GLAD
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Enable depth testing
    glEnable(GL_DEPTH_TEST);

    // Build and compile shaders
    const char* vertexShaderSource = R"(
    #version 330 core
    layout (location = 0) in vec3 aPos;      // Vertex position
    layout (location = 1) in vec3 aNormal;   // Vertex normal (not used)
    layout (location = 2) in vec3 aColor;    // Vertex color (for axes and grid)

    out vec3 vertexColor;  // Output color to the fragment shader

    uniform vec3 sphereColor;       // Uniform color for spheres
    uniform bool useUniformColor;   // Flag to determine which color to use

    uniform mat4 model;
    uniform mat4 view;
    uniform mat4 projection;

    void main()
    {
        if (useUniformColor)
            vertexColor = sphereColor;
        else
            vertexColor = aColor;
        gl_Position = projection * view * model * vec4(aPos, 1.0);
    }
    )";

    const char* fragmentShaderSource = R"(
    #version 330 core
    in vec3 vertexColor;  // Color from the vertex shader
    out vec4 FragColor;

    void main()
    {
        FragColor = vec4(vertexColor, 1.0);
    }
    )";

    unsigned int shaderProgram = createShaderProgram(vertexShaderSource, fragmentShaderSource);
    if (shaderProgram == 0)
    {
        std::cerr << "Shader program creation failed!" << std::endl;
        return -1;
    }

    // Get uniform locations
    unsigned int sphereColorLoc = glGetUniformLocation(shaderProgram, "sphereColor");
    unsigned int useUniformColorLoc = glGetUniformLocation(shaderProgram, "useUniformColor");

    // Generate sphere mesh
    std::vector<float> sphereVertices;
    std::vector<unsigned int> sphereIndices;
    float sphereRadius = 0.1f;
    unsigned int sectorCount = 36;
    unsigned int stackCount = 18;
    generateSphere(sphereVertices, sphereIndices, sphereRadius, sectorCount, stackCount);

    // Set up VAO and VBO for sphere
    unsigned int sphereVAO, sphereVBO, sphereEBO;

    glGenVertexArrays(1, &sphereVAO);
    glGenBuffers(1, &sphereVBO);
    glGenBuffers(1, &sphereEBO);

    glBindVertexArray(sphereVAO);

    glBindBuffer(GL_ARRAY_BUFFER, sphereVBO);
    glBufferData(GL_ARRAY_BUFFER, sphereVertices.size() * sizeof(float), &sphereVertices[0], GL_STATIC_DRAW);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sphereEBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sphereIndices.size() * sizeof(unsigned int), &sphereIndices[0], GL_STATIC_DRAW);

    // Position attribute (offset 0)
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // Normal attribute (offset 3 floats)
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(1);

    // Disable color attribute for spheres
    glDisableVertexAttribArray(2);

    glBindVertexArray(0);

    // Initialize empty spheres map
    std::map<std::tuple<int, int, int>, Sphere> spheres;

    // Initialize line vertices
    std::vector<float> lineVertices;
    unsigned int lineVAO, lineVBO;

    glGenVertexArrays(1, &lineVAO);
    glGenBuffers(1, &lineVBO);

    glBindVertexArray(lineVAO);

    glBindBuffer(GL_ARRAY_BUFFER, lineVBO);
    glBufferData(GL_ARRAY_BUFFER, 0, nullptr, GL_DYNAMIC_DRAW);

    // Position attribute
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // Color attribute
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(2);

    glBindVertexArray(0);

    // ------------------- Adding Axes and Grid Points ---------------------

    const float AXIS_LENGTH = 10.0f;

    std::vector<float> axesVertices = {
        // X-axis (Red)
        -AXIS_LENGTH, 0.0f, 0.0f,  1.0f, 0.0f, 0.0f,
         AXIS_LENGTH, 0.0f, 0.0f,  1.0f, 0.0f, 0.0f,
        // Y-axis (Green)
        0.0f, -AXIS_LENGTH, 0.0f,  0.0f, 1.0f, 0.0f,
        0.0f,  AXIS_LENGTH, 0.0f,  0.0f, 1.0f, 0.0f,
        // Z-axis (Blue)
        0.0f, 0.0f,  AXIS_LENGTH,  0.0f, 0.0f, 1.0f,
        0.0f, 0.0f, -AXIS_LENGTH,  0.0f, 0.0f, 1.0f,
    };

    // Generate grid points at integer coordinates
    std::vector<float> gridPoints;
    int gridRange = static_cast<int>(AXIS_LENGTH);

    for (int x = -gridRange; x <= gridRange; ++x)
    {
        for (int y = -gridRange; y <= gridRange; ++y)
        {
            for (int z = -gridRange; z <= gridRange; ++z)
            {
                gridPoints.push_back(static_cast<float>(x));
                gridPoints.push_back(static_cast<float>(y));
                gridPoints.push_back(static_cast<float>(z));

                // Color (white)
                gridPoints.push_back(1.0f); // R
                gridPoints.push_back(1.0f); // G
                gridPoints.push_back(1.0f); // B
            }
        }
    }

    unsigned int axesVAO, axesVBO;
    glGenVertexArrays(1, &axesVAO);
    glGenBuffers(1, &axesVBO);

    glBindVertexArray(axesVAO);
    glBindBuffer(GL_ARRAY_BUFFER, axesVBO);
    glBufferData(GL_ARRAY_BUFFER, axesVertices.size() * sizeof(float), axesVertices.data(), GL_STATIC_DRAW);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(2);
    glBindVertexArray(0);

    unsigned int gridVAO, gridVBO;
    glGenVertexArrays(1, &gridVAO);
    glGenBuffers(1, &gridVBO);

    glBindVertexArray(gridVAO);
    glBindBuffer(GL_ARRAY_BUFFER, gridVBO);
    glBufferData(GL_ARRAY_BUFFER, gridPoints.size() * sizeof(float), gridPoints.data(), GL_STATIC_DRAW);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(2);
    glBindVertexArray(0);

    // ------------------- Initialize ImGui ---------------------

    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;

    ImGui::StyleColorsLight();
    ImGuiStyle& style = ImGui::GetStyle();
    style.WindowRounding = 0.0f; // Remove window rounding
    style.Colors[ImGuiCol_WindowBg] = ImVec4(1.0f, 1.0f, 1.0f, 1.0f); // White background

    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init("#version 330");

    if (showUI)
    {
        glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
    }
    else
    {
        glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
    }

    while (!glfwWindowShouldClose(window))
    {
        float currentFrame = glfwGetTime();
        deltaTime = currentFrame - lastFrame;
        lastFrame = currentFrame;

        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();

        if (showUI)
        {
            if (glfwGetInputMode(window, GLFW_CURSOR) != GLFW_CURSOR_NORMAL)
            {
                glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
                firstMouse = true;
            }
        }
        else
        {
            if (glfwGetInputMode(window, GLFW_CURSOR) != GLFW_CURSOR_DISABLED)
            {
                glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
                firstMouse = true;
            }
            processInput(window);
        }

        if (showUI)
        {
            ImGui::SetNextWindowPos(ImVec2(0, 0));
            ImGui::SetNextWindowSize(ImVec2(static_cast<float>(SIDEBAR_WIDTH), static_cast<float>(SCR_HEIGHT)));
            ImGui::Begin("Control Panel", NULL, ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoCollapse);

            static int spherePos[3] = {0, 0, 0};
            ImGui::InputInt("X Position", &spherePos[0]);
            ImGui::InputInt("Y Position", &spherePos[1]);
            ImGui::InputInt("Z Position", &spherePos[2]);

            static bool highlight = false;
            ImGui::Checkbox("Hydrophobic", &highlight);

            if (ImGui::Button("Add Sphere"))
            {
                Sphere sphere;
                // The Z orientation seems to be reversed, hence the negative Z correction
                sphere.position = glm::vec3(spherePos[0], spherePos[1], -spherePos[2]);

                // Set color based on hydrophobic flag (highlight)
                if (highlight)
                    sphere.color = glm::vec3(1.0f, 0.0f, 0.0f);  // Red for hydrophobic
                else
                    sphere.color = glm::vec3(0.5f, 0.5f, 0.5f);  // Gray for non-hydrophobic

                // Add sphere to the map with its position as the key
                spheres[std::make_tuple(spherePos[0], spherePos[1], spherePos[2])] = sphere;

                // Update lines connecting the spheres
                updateLines(spheres, lineVertices, lineVBO);
            }

            ImGui::Checkbox("Show Grid Points", &showGridPoints);

            ImGui::InputInt("Sphere Size", &sphereSizeInt);
            if (sphereSizeInt < 1) sphereSizeInt = 1; // Ensure minimum value
            if (sphereSizeInt > 10) sphereSizeInt = 10; // Ensure maximum value
            sphereScale = static_cast<float>(sphereSizeInt);

            ImGui::InputInt("Line Thickness", &lineThicknessInt);
            if (lineThicknessInt < 1) lineThicknessInt = 1;
            if (lineThicknessInt > 10) lineThicknessInt = 10;
            lineThickness = static_cast<float>(lineThicknessInt);

            // Add JSON input field
            ImGui::Text("JSON Input:");
            ImGui::InputTextMultiline("##jsonInput", jsonInput, sizeof(jsonInput), ImVec2(-1.0f, 150.0f));

            if (ImGui::Button("Submit JSON"))
            {
                std::string jsonErrorMessage = "";
                std::string jsonString(jsonInput); // Convert char[] to std::string
                if (parseHPModelJson(jsonInput, spheres, jsonErrorMessage))
                {
                    // Update lines connecting the spheres
                    updateLines(spheres, lineVertices, lineVBO);
                }
                else
                {
                    // Parsing failed, error message is set
                }
            }

            // Display error message if parsing failed
            if (!jsonErrorMessage.empty())
            {
                ImGui::TextColored(ImVec4(1.0f, 0.0f, 0.0f, 1.0f), "%s", jsonErrorMessage.c_str());
            }

            ImGui::End();
        }

        glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Use the shader program
        glUseProgram(shaderProgram);

        glm::mat4 view = glm::lookAt(cameraPos, cameraPos + cameraFront, cameraUp);
        glm::mat4 projection = glm::perspective(glm::radians(fov), static_cast<float>(SCR_WIDTH) / static_cast<float>(SCR_HEIGHT), 0.1f, 100.0f);

        unsigned int modelLoc = glGetUniformLocation(shaderProgram, "model");
        unsigned int viewLoc  = glGetUniformLocation(shaderProgram, "view");
        unsigned int projLoc  = glGetUniformLocation(shaderProgram, "projection");

        glUniformMatrix4fv(viewLoc, 1, GL_FALSE, glm::value_ptr(view));
        glUniformMatrix4fv(projLoc, 1, GL_FALSE, glm::value_ptr(projection));

        // Render axis labels
        ImDrawList* drawList = ImGui::GetBackgroundDrawList(); // Or ImGui::GetForegroundDrawList()

        ImU32 textColor = IM_COL32(255, 255, 255, 255); // White color

        ImVec2 screenPos;
        if (WorldToScreen(glm::vec3(AXIS_LENGTH, 0.0f, 0.0f), view, projection, screenPos))
        {
            drawList->AddText(screenPos, textColor, "X+");
        }
        if (WorldToScreen(glm::vec3(-AXIS_LENGTH, 0.0f, 0.0f), view, projection, screenPos))
        {
            drawList->AddText(screenPos, textColor, "X-");
        }
        if (WorldToScreen(glm::vec3(0.0f, AXIS_LENGTH, 0.f), view, projection, screenPos))
        {
            drawList->AddText(screenPos, textColor, "Y+");
        }
        if (WorldToScreen(glm::vec3(0.0f, -AXIS_LENGTH, 0.0f), view, projection, screenPos))
        {
            drawList->AddText(screenPos, textColor, "Y-");
        }
        if (WorldToScreen(glm::vec3(0.0f, 0.0f, AXIS_LENGTH), view, projection, screenPos))
        {
            drawList->AddText(screenPos, textColor, "Z-");
        }
        if (WorldToScreen(glm::vec3(0.0f, 0.0f, -AXIS_LENGTH), view, projection, screenPos))
        {
            drawList->AddText(screenPos, textColor, "Z+");
        }

        // **Use per-vertex colors for axes and grid**
        glUniform1i(useUniformColorLoc, 0);  // false

        // Draw axes
        glBindVertexArray(axesVAO);
        glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(glm::mat4(1.0f)));
        glDrawArrays(GL_LINES, 0, 6);
        glBindVertexArray(0);

        // Draw grid points if enabled
        if (showGridPoints)
        {
            glBindVertexArray(gridVAO);
            glPointSize(5.0f);
            glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(glm::mat4(1.0f)));
            glDrawArrays(GL_POINTS, 0, static_cast<unsigned int>(gridPoints.size() / 6));
            glBindVertexArray(0);
        }

        // **Use uniform color for spheres**
        glUniform1i(useUniformColorLoc, 1);  // true

        glBindVertexArray(sphereVAO);
        for (const auto& [key, sphere] : spheres)
        {
            glm::mat4 model = glm::mat4(1.0f);
            model = glm::translate(model, sphere.position);
            model = glm::scale(model, glm::vec3(sphereScale)); // Scale each sphere

            glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(model));

            // Set the sphere color via uniform
            glUniform3f(sphereColorLoc, sphere.color.r, sphere.color.g, sphere.color.b);

            glDrawElements(GL_TRIANGLES, static_cast<unsigned int>(sphereIndices.size()), GL_UNSIGNED_INT, 0);
        }
        glBindVertexArray(0);

        // **Use per-vertex colors for lines**
        glUniform1i(useUniformColorLoc, 0);  // false

        glLineWidth(lineThickness); // Use the variable lineThickness
        glBindVertexArray(lineVAO);
        glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(glm::mat4(1.0f)));
        glDrawArrays(GL_LINES, 0, static_cast<unsigned int>(lineVertices.size() / 6));
        glBindVertexArray(0);
        glLineWidth(1.0f); // Reset to default after drawing

        updateWindowTitle(window, showUI);

        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        glfwSwapBuffers(window);
        glfwPollEvents();
    }

    glDeleteVertexArrays(1, &sphereVAO);
    glDeleteBuffers(1, &sphereVBO);
    glDeleteBuffers(1, &sphereEBO);

    glDeleteVertexArrays(1, &lineVAO);
    glDeleteBuffers(1, &lineVBO);

    glDeleteVertexArrays(1, &axesVAO);
    glDeleteBuffers(1, &axesVBO);

    glDeleteVertexArrays(1, &gridVAO);
    glDeleteBuffers(1, &gridVBO);

    glDeleteProgram(shaderProgram);

    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwTerminate();

    return 0;
}

void generateSphere(std::vector<float>& vertices, std::vector<unsigned int>& indices, float radius, unsigned int sectorCount, unsigned int stackCount)
{
    float x, y, z, xy;
    float nx, ny, nz, lengthInv = 1.0f / radius;

    float sectorStep = 2 * glm::pi<float>() / sectorCount;
    float stackStep = glm::pi<float>() / stackCount;
    float sectorAngle, stackAngle;

    for(unsigned int i = 0; i <= stackCount; ++i)
    {
        stackAngle = glm::pi<float>() / 2 - i * stackStep;
        xy = radius * cosf(stackAngle);
        z = radius * sinf(stackAngle);

        for(unsigned int j = 0; j <= sectorCount; ++j)
        {
            sectorAngle = j * sectorStep;

            x = xy * cosf(sectorAngle);
            y = xy * sinf(sectorAngle);

            nx = x * lengthInv;
            ny = y * lengthInv;
            nz = z * lengthInv;

            vertices.push_back(x);
            vertices.push_back(y);
            vertices.push_back(z);

            vertices.push_back(nx);
            vertices.push_back(ny);
            vertices.push_back(nz);
        }
    }

    unsigned int k1, k2;
    for(unsigned int i = 0; i < stackCount; ++i)
    {
        k1 = i * (sectorCount + 1);
        k2 = k1 + sectorCount + 1;

        for(unsigned int j = 0; j < sectorCount; ++j, ++k1, ++k2)
        {
            if(i != 0)
            {
                indices.push_back(k1);
                indices.push_back(k2);
                indices.push_back(k1 + 1);
            }

            if(i != (stackCount-1))
            {
                indices.push_back(k1 + 1);
                indices.push_back(k2);
                indices.push_back(k2 + 1);
            }
        }
    }
}

void updateLines(const std::map<std::tuple<int, int, int>, Sphere>& spheres,
                std::vector<float>& lineVertices, unsigned int lineVBO)
{
    lineVertices.clear();

    for (const auto& [key, sphere] : spheres)
    {
        int x = std::get<0>(key);
        int y = std::get<1>(key);
        int z = std::get<2>(key);

        for (int dx = -1; dx <= 1; ++dx)
        {
            for (int dy = -1; dy <= 1; ++dy)
            {
                for (int dz = -1; dz <= 1; ++dz)
                {
                    if ((abs(dx) + abs(dy) + abs(dz)) != 1)
                        continue;

                    int nx = x + dx;
                    int ny = y + dy;
                    int nz = z + dz;

                    auto neighborKey = std::make_tuple(nx, ny, nz);
                    if (spheres.find(neighborKey) != spheres.end())
                    {
                        if (key < neighborKey)
                        {
                            glm::vec3 neighborPos = spheres.at(neighborKey).position;

                            lineVertices.insert(lineVertices.end(), {
                                sphere.position.x, sphere.position.y, sphere.position.z, 0.5f, 0.5f, 0.5f,
                                neighborPos.x, neighborPos.y, neighborPos.z, 0.5f, 0.5f, 0.5f
                            });
                        }
                    }
                }
            }
        }
    }

    glBindBuffer(GL_ARRAY_BUFFER, lineVBO);
    glBufferData(GL_ARRAY_BUFFER, lineVertices.size() * sizeof(float), &lineVertices[0], GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void processInput(GLFWwindow *window)
{
    float cameraSpeed = 10.0f * deltaTime;

    if (glfwGetKey(window, GLFW_KEY_W) == GLFW_PRESS)
        cameraPos += cameraSpeed * cameraFront;
    if (glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS)
        cameraPos -= cameraSpeed * cameraFront;
    if (glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS)
        cameraPos -= glm::normalize(glm::cross(cameraFront, cameraUp)) * cameraSpeed;
    if (glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS)
        cameraPos += glm::normalize(glm::cross(cameraFront, cameraUp)) * cameraSpeed;
    if (glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS)
        cameraPos += cameraSpeed * cameraUp;
    if (glfwGetKey(window, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS)
        cameraPos -= cameraSpeed * cameraUp;

    if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS)
        glfwSetWindowShouldClose(window, true);
}

void framebuffer_size_callback(GLFWwindow* window, int width, int height)
{
    SCR_WIDTH = width;
    SCR_HEIGHT = height;
    glViewport(0, 0, SCR_WIDTH, SCR_HEIGHT);
}

void mouse_callback(GLFWwindow* window, double xposIn, double yposIn)
{
    if (glfwGetInputMode(window, GLFW_CURSOR) != GLFW_CURSOR_DISABLED)
        return;

    float xpos = static_cast<float>(xposIn);
    float ypos = static_cast<float>(yposIn);

    if (firstMouse)
    {
        lastX = xpos;
        lastY = ypos;
        firstMouse = false;
    }

    float xoffset = xpos - lastX;
    float yoffset = lastY - ypos;

    lastX = xpos;
    lastY = ypos;

    float sensitivity = 0.1f;
    xoffset *= sensitivity;
    yoffset *= sensitivity;

    yaw   += xoffset;
    pitch += yoffset;

    if(pitch > 89.0f)
        pitch = 89.0f;
    if(pitch < -89.0f)
        pitch = -89.0f;

    glm::vec3 front;
    front.x = cos(glm::radians(yaw)) * cos(glm::radians(pitch));
    front.y = sin(glm::radians(pitch));
    front.z = sin(glm::radians(yaw)) * cos(glm::radians(pitch));

    cameraFront = glm::normalize(front);
}

void scroll_callback(GLFWwindow* window, double xoffset, double yoffset)
{
    fov -= static_cast<float>(yoffset);
    if(fov < 1.0f)
        fov = 1.0f;
    if(fov > 90.0f)
        fov = 90.0f;
}

unsigned int createShaderProgram(const char* vertexShaderSource, const char* fragmentShaderSource)
{
    unsigned int vertexShader;
    vertexShader = glCreateShader(GL_VERTEX_SHADER);

    glShaderSource(vertexShader, 1, &vertexShaderSource, NULL);
    glCompileShader(vertexShader);

    int success;
    char infoLog[512];

    glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &success);
    if(!success)
    {
        glGetShaderInfoLog(vertexShader, 512, NULL, infoLog);
        std::cout << "ERROR::SHADER::VERTEX::COMPILATION_FAILED\n"
                  << infoLog << std::endl;
        return 0;
    }

    unsigned int fragmentShader;
    fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);

    glShaderSource(fragmentShader, 1, &fragmentShaderSource, NULL);
    glCompileShader(fragmentShader);

    glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &success);
    if(!success)
    {
        glGetShaderInfoLog(fragmentShader, 512, NULL, infoLog);
        std::cout << "ERROR::SHADER::FRAGMENT::COMPILATION_FAILED\n"
                  << infoLog << std::endl;
        return 0;
    }

    unsigned int shaderProgram;
    shaderProgram = glCreateProgram();

    glAttachShader(shaderProgram, vertexShader);
    glAttachShader(shaderProgram, fragmentShader);
    glLinkProgram(shaderProgram);

    glGetProgramiv(shaderProgram, GL_LINK_STATUS, &success);
    if(!success)
    {
        glGetProgramInfoLog(shaderProgram, 512, NULL, infoLog);
        std::cout << "ERROR::SHADER::PROGRAM::LINKING_FAILED\n"
                  << infoLog << std::endl;
        return 0;
    }

    glDeleteShader(vertexShader);
    glDeleteShader(fragmentShader);

    return shaderProgram;
}

void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
    if (key == GLFW_KEY_TAB && action == GLFW_PRESS)
    {
        showUI = !showUI;
        if (showUI)
        {
            glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
        }
        else
        {
            glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
            firstMouse = true;
        }
    }
}

void updateWindowTitle(GLFWwindow* window, bool showUI)
{
    std::string title = "3D Spheres with Connections - ";
    title += showUI ? "UI Interaction Mode" : "Camera Control Mode";
    glfwSetWindowTitle(window, title.c_str());
}

bool WorldToScreen(const glm::vec3& worldPos, const glm::mat4& view, const glm::mat4& projection, ImVec2& screenPos)
{
    glm::vec4 clipSpacePos = projection * view * glm::vec4(worldPos, 1.0f);

    if (clipSpacePos.w <= 0.0f)
        return false; // Behind the camera

    glm::vec3 ndc = glm::vec3(clipSpacePos) / clipSpacePos.w;

    if (ndc.x < -1.0f || ndc.x > 1.0f || ndc.y < -1.0f || ndc.y > 1.0f)
        return false; // Outside of screen

    // Map to window coordinates
    float windowX = (ndc.x + 1.0f) * SCR_WIDTH / 2.0f;
    float windowY = (1.0f - ndc.y) * SCR_HEIGHT / 2.0f; // Invert Y for window coordinates

    screenPos = ImVec2(windowX, windowY);

    return true;
}

bool parseHPModelJson(const std::string& jsonInput, std::map<std::tuple<int, int, int>, Sphere>& spheres, std::string& errorMessage)
{
    // Remove all whitespace
    std::string input;
    for (char c : jsonInput)
    {
        if (!isspace(static_cast<unsigned char>(c)))
            input += c;
    }

    if (input.empty())
    {
        errorMessage = "Input is empty.";
        return false;
    }

    if (input.front() != '[' || input.back() != ']')
    {
        errorMessage = "JSON must start with '[' and end with ']'.";
        return false;
    }

    input = input.substr(1, input.size() - 2); // Remove '[' and ']'

    size_t pos = 0;
    while (pos < input.size())
    {
        if (input[pos] != '{')
        {
            errorMessage = "Expected '{' at position " + std::to_string(pos);
            return false;
        }

        size_t endPos = input.find('}', pos);
        if (endPos == std::string::npos)
        {
            errorMessage = "Expected '}' after position " + std::to_string(pos);
            return false;
        }

        std::string objStr = input.substr(pos + 1, endPos - pos - 1); // Get content inside '{' and '}'

        // Parse the object string
        int x = 0, y = 0, z = 0;
        bool hydrophobic = false;
        bool xSet = false, ySet = false, zSet = false, hydroSet = false;

        size_t objPos = 0;
        while (objPos < objStr.size())
        {
            // Find key
            if (objStr[objPos] != '"')
            {
                errorMessage = "Expected '\"' for key at position " + std::to_string(objPos);
                return false;
            }
            size_t keyEnd = objStr.find('"', objPos + 1);
            if (keyEnd == std::string::npos)
            {
                errorMessage = "Expected '\"' to end key at position " + std::to_string(objPos + 1);
                return false;
            }

            std::string key = objStr.substr(objPos + 1, keyEnd - objPos - 1);

            // Find ':'
            size_t colonPos = objStr.find(':', keyEnd + 1);
            if (colonPos == std::string::npos)
            {
                errorMessage = "Expected ':' after key \"" + key + "\"";
                return false;
            }

            // Find value
            size_t valueStart = colonPos + 1;
            size_t valueEnd = valueStart;

            if (key == "hydrophobic")
            {
                // Value should be 'true' or 'false'
                if (objStr.compare(valueStart, 4, "true") == 0)
                {
                    hydrophobic = true;
                    valueEnd = valueStart + 4;
                }
                else if (objStr.compare(valueStart, 5, "false") == 0)
                {
                    hydrophobic = false;
                    valueEnd = valueStart + 5;
                }
                else
                {
                    errorMessage = "Invalid value for key \"hydrophobic\"";
                    return false;
                }
                hydroSet = true;
            }
            else
            {
                // Value should be an integer
                size_t numEnd = valueStart;
                if (objStr[numEnd] == '-') numEnd++; // Handle negative numbers
                while (numEnd < objStr.size() && isdigit(objStr[numEnd]))
                {
                    numEnd++;
                }
                if (numEnd == valueStart)
                {
                    errorMessage = "Invalid integer value for key \"" + key + "\"";
                    return false;
                }
                std::string valueStr = objStr.substr(valueStart, numEnd - valueStart);
                int value = 0;
                try
                {
                    value = std::stoi(valueStr);
                }
                catch (std::exception&)
                {
                    errorMessage = "Invalid integer value for key \"" + key + "\"";
                    return false;
                }
                if (key == "x")
                {
                    x = value;
                    xSet = true;
                }
                else if (key == "y")
                {
                    y = value;
                    ySet = true;
                }
                else if (key == "z")
                {
                    z = value;
                    zSet = true;
                }
                else
                {
                    errorMessage = "Unexpected key \"" + key + "\"";
                    return false;
                }
                valueEnd = numEnd;
            }

            // Skip commas
            objPos = valueEnd;
            if (objPos < objStr.size() && objStr[objPos] == ',')
            {
                objPos++;
            }
        }

        if (!xSet || !ySet || !zSet || !hydroSet)
        {
            errorMessage = "Missing key(s) in object.";
            return false;
        }

        // Add sphere
        Sphere sphere;
        sphere.position = glm::vec3(x, y, -z); // Adjust Z-axis as in your code
        sphere.color = hydrophobic ? glm::vec3(1.0f, 0.0f, 0.0f) : glm::vec3(0.5f, 0.5f, 0.5f);

        spheres[std::make_tuple(x, y, z)] = sphere;

        // Move pos to after the '}'
        pos = endPos + 1;

        // Skip commas
        if (pos < input.size() && input[pos] == ',')
        {
            pos++;
        }
    }

    return true;
}
